@@ -21,19 +21,21 @@ import numpy as np
 import yaml
 import pyparsing as pp
 import re
+import pandas as pd
 # to make it compatible with Python >= 3.7
 try:
     import SchemDraw as schemdraw
     import SchemDraw.elements as elm
 except ImportError:
     import schemdraw
-    import schemdraw.elements as elm
+    import schemdraw.elements.legacy as elm
 
+from scipy.integrate import simps
 from collections import Counter
 from scipy.constants import epsilon_0 as e0
-from .elements import Z_C, Z_stray, log, parallel, Z_R, Z_L, Z_w, Z_ws, Z_wo
-from .loss import Z_in, Z_loss
-from .cole_cole import cole_cole_model, cole_cole_R_model, cole_cole_2_model, cole_cole_3_model, cole_cole_4_model, havriliak_negami, cole_cole_2tissue_model, havriliak_negamitissue
+from .elements import Z_C, Z_stray, log, parallel, Z_R, Z_L, Z_w, Z_ws, Z_wo, eps
+from .loss import Z_in, Z_loss, Z_skin
+from .cole_cole import cole_cole_model, cole_cole_R_model, cole_cole_2_model, cole_cole_3_model, cole_cole_4_model, havriliak_negami, cole_cole_2tissue_model, havriliak_negamitissue, raicu
 from .double_shell import double_shell_model
 from .randles import Z_randles, Z_randles_CPE
 from .RC import RC_model, rc_model, drc_model, rc_tau_model
@@ -41,11 +43,52 @@ from .cpe import cpe_model, cpe_ct_model, cpe_ct_w_model, cpe_onset_model, cpeti
 from .single_shell import single_shell_model
 from lmfit import Model, CompositeModel
 from copy import deepcopy
-from . import logger
 from packaging import version
+import logging
+
+logger = logging.getLogger(__name__)
 
 
-def convert_diel_properties_to_impedance(omega, eps, sigma, c0):
+def convert_diel_properties_to_impedance(omega, eps_r, sigma, c0):
+    r"""Return impedance from dielectric properties.
+
+    Parameters
+    ----------
+
+    omega: :class:`numpy.ndarray`, double
+        frequency array
+    eps_r: :class:`numpy.ndarray`, double
+        relative permittivity
+    sigma: :class:`numpy.ndarray`, double
+        conductivity in S/m
+    c0: double
+        unit capacitance of device
+
+    Returns
+    -------
+
+    :class:`numpy.ndarray`, complex
+        impedance array
+
+    Notes
+    -----
+
+    Use that the impedance is
+
+    .. math::
+
+        Z = (j \omega \varepsilon_\mathrm{r}^\ast c_0)^{-1} ,
+
+    where :math:`\varepsilon_\mathrm{r}^\ast` is the relative complex permittivity (see for instance [Grant1958]_
+    for further explanation). Note that the vacuum permittivity :math:`\varepsilon_0` is contained in :math:`c_0`.
+
+    In the function, the variable `epsc` describes the term
+
+    .. math::
+
+        \omega \varepsilon_\mathrm{r}^\ast
+
+    """
     epsc = omega * eps - 1j * sigma / e0
     return 1. / (1j * epsc * c0)
 
@@ -61,10 +104,10 @@ def return_diel_properties(omega, Z, c0):
 
     .. math::
 
-        Z = (j \omega \varepsilon^\ast c_0)^{-1} ,
+        Z = (j \omega \varepsilon_\mathrm{r}^\ast c_0)^{-1} ,
 
-    where :math:`\varepsilon^\ast` is the complex permittivity (see for instance [Grant1958]_
-    for further explanation).
+    where :math:`\varepsilon_\mathrm{r}^\ast` is the relative complex permittivity (see for instance [Grant1958]_
+    for further explanation). Note that the vacuum permittivity :math:`\varepsilon_0` is contained in :math:`c_0`.
 
     When the unit capacitance :math:`c_0` of the device is known, a direct mapping from impedance to
     relative complex permittivity is possible:
@@ -116,6 +159,47 @@ def return_diel_properties(omega, Z, c0):
     return eps_r, conductivity
 
 
+def return_dielectric_modulus(omega, Z, c0):
+    r"""Return dielectric modulus
+
+    Notes
+    -----
+
+    The dielectric modulus is  :math:`M = 1 / \varepsilon_\mathrm{r}^\ast`.
+    See [Bordi2001]_ for further explanation.
+
+    Parameters
+    ----------
+
+    omega: :class:`numpy.ndarray`, double
+        frequency array
+    Z: :class:`numpy.ndarray`, complex
+        impedance array
+    c0: double
+        unit capacitance of device
+
+    Returns
+    -------
+
+    ReM: :class:`numpy.ndarray`, double
+        real part of modulus
+    ImM: :class:`numpy.ndarray`, double
+        imaginary part of modulus
+
+    References
+    ----------
+
+    .. [Bordi2001] Bordi, F., & Cametti, C. (2001). Occurrence of an intermediate relaxation process in water-in-oil microemulsions below percolation: The electrical modulus formalism.
+                   Journal of Colloid and Interface Science, 237(2), 224–229.
+                   https://doi.org/10.1006/jcis.2001.7456
+    """
+    epsc = 1. / (1j * omega * Z * c0)
+    M = 1. / epsc
+    ReM = M.real
+    ImM = M.imag
+    return ReM, ImM
+
+
 def check_parameters(bufdict):
     """Check parameters for physical correctness.
 
@@ -125,6 +209,11 @@ def check_parameters(bufdict):
     bufdict: dict
         Contains all parameters and their values
 
+    Notes
+    -----
+
+    All parameters are forced to be greater or equal zero.
+    There are only two exceptions.
     """
 
     # capacitances in pF
@@ -132,7 +221,7 @@ def check_parameters(bufdict):
     capacitances = ['C']
     # taus in ns
     taus = ['tau', 'tauE', 'tau1', 'tau2', 'tau3', 'tau4']
-    zerotoones = ['p', 'a', 'alpha', 'beta', 'a1', 'a2', 'a3', 'a4', 'nu']
+    zerotoones = ['p', 'a', 'alpha', 'beta', 'gamma', 'a1', 'a2', 'a3', 'a4', 'nu']
     permittivities = ['em', 'ecp', 'emed', 'ene', 'enp', 'epsl', 'eps', 'epsinf']
 
     # __lnsigma and Rk (Lin-KK test)
@@ -165,11 +254,11 @@ def check_parameters(bufdict):
                    Change your initial value.""".format(p)
             if bufdict[p].vary:
                 if bufdict[p].min < 0.0 or bufdict[p].min > 1.0:
-                    logger.info("""{} is an exponent that needs to be between
+                    logger.debug("""{} is an exponent that needs to be between
                                    0.0 and 1.0. Changed your min value to 0.""".format(p))
                     bufdict[p].set(min=0.0)
                 if bufdict[p].max > 1.0:
-                    logger.info("""{} is an exponent that needs to be between 0.0 and 1.0.
+                    logger.debug("""{} is an exponent that needs to be between 0.0 and 1.0.
                                    Changed your max value to 1.0.""".format(p))
                     bufdict[p].set(max=1.0)
             continue
@@ -185,11 +274,11 @@ def check_parameters(bufdict):
                    or equal to 1. Change the initial value.""".format(p)
             if bufdict[p].vary:
                 if bufdict[p].min < 1.0:
-                    logger.info("""The permittivity {} needs to be greater
+                    logger.debug("""The permittivity {} needs to be greater
                                    than or equal to 1. Changed the min value to 1.""".format(p))
                     bufdict[p].set(min=1.0)
                 if bufdict[p].max < 1.0:
-                    logger.info("""The permittivity {} needs to be greater than
+                    logger.debug("""The permittivity {} needs to be greater than
                                    or equal to 1. Changed the max value to inf.""".format(p))
                     bufdict[p].set(max=np.inf)
             continue
@@ -207,7 +296,7 @@ def check_parameters(bufdict):
     return bufdict
 
 
-def set_parameters(model, parameterdict=None, emcee=False):
+def set_parameters(model, parameterdict=None, emcee=False, weighting_model=False):
     """
     Parameters
     -----------
@@ -219,6 +308,8 @@ def set_parameters(model, parameterdict=None, emcee=False):
         If it is None (default), the parameters are read in from a yaml-file.
     emcee: bool, optional
         if emcee is used, an additional `__lnsigma` parameter will be set
+    weighting_model: bool, optional
+        if a weighting model is used, the variance will be fit as well
 
 
     Returns
@@ -257,6 +348,23 @@ def set_parameters(model, parameterdict=None, emcee=False):
     elif emcee and "__lnsigma" in parameterdict:
         parameters.add("__lnsigma", **parameterdict["__lnsigma"])
 
+    if weighting_model:
+        if "stdA" not in parameterdict and "stdPhi" not in parameterdict:
+            raise RuntimeError("You need to provide the variables stdA and stdPhi if you want to use a weighting model.")
+        for var in ["stdA", "stdPhi"]:
+            if "min" in parameterdict[var]:
+                if parameterdict[var]["min"] < 0:
+                    logger.warning("You set the minimum value of parameter {} to a negative value. That does not work and the value is set to 0.".format(var))
+                    parameterdict[var]["min"] = 0
+            else:
+                parameterdict[var]["min"] = 0
+            if "max" in parameterdict[var]:
+                if parameterdict[var]["max"] < 0:
+                    logger.warning("You set the maximum value of parameter {} to a negative value. That does not work and the value is set to inf.".format(var))
+                parameterdict[var]["max"] = np.inf
+            else:
+                parameterdict[var]["min"] = 0
+            parameters.add(var, **parameterdict[var])
     return parameters
 
 
@@ -375,6 +483,7 @@ def available_models():
               'ColeCole3',
               'ColeCole2',
               'HavriliakNegami',
+              'Raicu',
               'Randles',
               'RandlesCPE',
               'RCfull',
@@ -393,6 +502,7 @@ def available_models():
               'Wo',
               'LCR',
               'LR',
+              'LRSkin',
               'Ws',
               'Cstray']
     return models
@@ -461,6 +571,8 @@ def _model_function(modelname):
         model = havriliak_negami
     elif modelname == 'HavriliakNegamiTissue':
         model = havriliak_negamitissue
+    elif modelname == "Raicu":
+        model = raicu
     elif modelname == 'ColeColeR':
         model = cole_cole_R_model
     elif modelname == 'Randles':
@@ -507,6 +619,8 @@ def _model_function(modelname):
         model = Z_loss
     elif modelname == "LR":
         model = Z_in
+    elif modelname == "LRSkin":
+        model = Z_skin
     elif modelname == "Cstray":
         model = Z_stray
     else:
@@ -658,10 +772,18 @@ def _check_circuit(circuit, startpar=False):
             else:
                 raise RuntimeError("You must have entered a wrong circuit!")
         elif isinstance(c, list):
+            # we only have a list if there is any parallel circuit in the circuit
+            if circuit.count(",") != 1:
+                # if we start with a parallel element, the first entry in the list must contain a comma
+                if startpar and circuit[0].count(",") != 1:
+                    raise RuntimeError("You must have entered a wrong circuit! There is a comma missing.")
+                # otherwise there must be one comma somewhere in the circuit
+                if c.count(",") != 1:
+                    raise RuntimeError("You must have entered a wrong circuit! There is a comma missing.")
             _check_circuit(c)
 
 
-def get_equivalent_circuit_model(modelname, logscale=False):
+def get_equivalent_circuit_model(modelname, logscale=False, diel=False):
     """Get LMFIT CompositeModel.
 
     Parameters
@@ -670,6 +792,8 @@ def get_equivalent_circuit_model(modelname, logscale=False):
         String representation of the equivalent circuit.
     logscale: bool
         Convert to logscale.
+    diel: bool
+        Convert to complex permittivity and fit this instead of impedance.
 
     Returns
     -------
@@ -701,14 +825,40 @@ def get_equivalent_circuit_model(modelname, logscale=False):
     circuit = _process_circuit(circuitstr.asList()[0])
     if logscale:
         circuit = CompositeModel(circuit, Model(dummy), log)
+    elif diel:
+        circuit = CompositeModel(circuit, Model(make_eps), eps)
+    if logscale and diel:
+        raise RuntimeError("You must chose the representation of the impedance value")
+    _check_models_suffix(circuit)
     logger.debug("Created composite model {}".format(circuit))
     return circuit
 
 
+def _check_models_suffix(circuit):
+    baseparams = []
+    for p in circuit.param_names:
+        tmp = p.split("_")
+        suf = None
+        if len(tmp) == 2:
+            suf = tmp[0]
+            par = tmp[1]
+        elif len(tmp) == 1:
+            par = tmp[0]
+        else:
+            raise RuntimeError("The parameter {} cannot be split in prefix and suffix.".format(p))
+        if par in baseparams and suf is None:
+            raise RuntimeError("The parameter {} has no prefix (its model has no suffix). This will lead to wrong parameter assignments. Change the part of the model, to which this parameter belongs (add a unique suffix).".format(p))
+        baseparams.append(par)
+
+
 def dummy(omega):
     return np.ones(omega.shape)
 
 
+def make_eps(omega, c0all):
+    return 1. / (1j * omega * c0all)
+
+
 def _model_label(model):
     labels = {'ColeCole': 'Cole-Cole',
               'ColeColeR': 'Cole-Cole w/ R',
@@ -716,6 +866,7 @@ def _model_label(model):
               'ColeCole3': '3 Cole-Cole',
               'ColeCole2': '2 Cole-Cole',
               'HavriliakNegami': 'Havriliak-Negami',
+              'Raicu': 'Raicu',
               'Randles': 'Randles',
               'RandlesCPE': 'Randles w/ CPE',
               'RCfull': 'RC',
@@ -723,8 +874,10 @@ def _model_label(model):
               'SingleShell': 'Single Shell',
               'DoubleShell': 'DoubleShell',
               'CPE': 'CPE',
+              'CPETissue': 'CPE for tissues',
               'CPEonset': 'CPEonset',
               'CPECT': 'CPECT',
+              'CPECTTissue': 'CPECT for tissues',
               'CPECTW': 'CPECTW',
               'DRC': 'DRC',
               'L': 'L',
@@ -752,6 +905,7 @@ def _get_element(name):
                     'ColeCole4',
                     'ColeCole3',
                     'ColeCole2',
+                    'Raicu',
                     'HavriliakNegami',
                     'Randles',
                     'RandlesCPE',
@@ -762,6 +916,8 @@ def _get_element(name):
                     'DoubleShell']
     capacitorlike = ['CPE',
                      'CPEonset',
+                     'CPETissue',
+                     'CPECTTissue',
                      'CPECT',
                      'CPECTW',
                      'W',
@@ -1001,3 +1157,74 @@ def _get_step_width(circuit, distance):
             counter += 1
     step = distance / counter
     return step
+
+
+def _return_resistance_capacitance(omega, Z):
+    R = Z.real
+    C = -1.0 / (omega * Z.imag)
+    return R, C
+
+
+def KK_integral_transform(omega, Z):
+    """Kramers-Kronig integral transform.
+
+    Parameters
+    ----------
+    omega: :class:`numpy.ndarray`, double
+        frequency array
+    Z: :class:`numpy.ndarray`, complex
+        impedance array
+
+    Returns
+    -------
+    :class:`numpy.ndarray`, complex
+        The transformed impedance array.
+
+    Notes
+    -----
+
+    Implementation following [Urquidi1990]_
+
+    .. [Urquidi1990] Urquidi-Macdonald, M., Real, S., & Macdonald, D. D. (1990).
+                     Applications of Kramers-Kronig transforms in the analysis of electrochemical impedance data-III. Stability and linearity.
+                     Electrochimica Acta, 35(10), 1559–1566.
+                     https://doi.org/10.1016/0013-4686(90)80010-L
+    """
+
+    ZKK = np.ndarray(omega.shape, dtype=complex)
+    for i, w in enumerate(omega):
+        x = np.append(omega[:i], omega[i + 1:])
+        real = np.append(Z.real[:i], Z.real[i + 1:])
+        imag = np.append(Z.imag[:i], Z.imag[i + 1:])
+        # real part
+        integrand = (x * imag - w * Z.imag[i]) / (x * x - w * w)
+        ZKK[i] = -2. / np.pi * simps(integrand, x=x)
+        # imag part
+        integrand = (real - Z.real[i]) / (x * x - w * w)
+        ZKK[i] += 1j * 2. * w / np.pi * simps(integrand, x=x)
+    return ZKK
+
+
+def save_impedance(omega, impedance, format="CSV", filename="impedance"):
+    """Save impedance to CSV or XLSX file.
+
+    Parameters
+    ----------
+    omega: :class:`numpy.ndarray`, double
+        frequency array
+    impedance: :class:`numpy.ndarray`, complex
+        impedance array
+    format: str
+        use either CSV or XLSX format. Based on the format, the correct ending is chosen.
+    filename: str
+        specify a filename (without ending!). the default is impedance.csv or impedance.xlsx
+    """
+    assert isinstance(filename, str), "You need to provide a str as filename!"
+    outdict = {"freq": omega / (2. * np.pi), 'real': impedance.real, 'imag': impedance.imag}
+    df = pd.DataFrame(outdict)
+    if format == "CSV":
+        df.to_csv(filename + ".csv", index=False)
+    elif format == "XLSX":
+        df.to_excel(filename + ".xlsx", index=False)
+    else:
+        raise(RuntimeError("You provided a wrong format. Use either CSV or XLSX."))
